# -*- coding: utf-8 -*-
# Tässä pelissä läiskitään ympäriinsä lenteleviä hyttysiä

# Ohjelmassa tarvitaan näitä python-moduleja
import pygame
import random
from math import sqrt  # math-modulista tarvitaan vain sqrt-funktio
import OhjeModuli   # Itse tekemäni moduli

# rgb-koodattuja värejä.
Musta = (0, 0, 0)  # Mustassa ei ole mitään valoa
Sin = (0, 0, 255)  # vain sinistä
Pun = (255, 0, 0)
Vihr = (0, 255, 0)
Valk = (255, 255, 255)  # valkoisessa on kaikenväristä valoa
Tausta = (160, 200, 255)

# Päivitetään näyttö näin monta kertaa sekunnissa.
# Pelisilmukka suoritetaan näin monta kertaa sekunnissa
Nayttotaajuus = 24  # Voit kokeilla eri näyttötaajuuksia
Dt = 1.0/Nayttotaajuus

# Peliin liittyviä muuttujia.
VLatka = 200.0  # Lätkän nopeus
Vmax = 250.0
MaxKiihtyvyys = 400.0
N_hyttyset = 100


# Hyttyslätkään liittyvät muuttujat ja metodit
# __init__()-funktio antaa alkuarvot muuttujille.
# Niitä muutetaan pelin kuluessa.
class Cl_Latka:
    def __init__(self, x, y, koko):
        self.paikka = (x, y)  # Paikka näytöllä
        self.koko = koko   # Halkaisija
        self.color = Vihr
        self.liikkuu = False

# Metodi, joka piirtää lätkän näytölle aina kutsuttaessa
# pygame kirjasto haluaa koordinaatit pikseleinä, siksi x ja y
# muutetaan kokonaisluvuiksi int-funktiolla
    def piirra(self):
        (x, y) = self.paikka
        iPaikka = (int(x), int(y))
        pygame.draw.circle(screen, self.color, iPaikka, self.koko, 0)
        pygame.draw.circle(screen, Pun, iPaikka, int(self.koko/2), 0)

# Tämä metodi siirtää lätkää nopeudella (vx, vy)
    def liiku(self, vxvy):
        (x, y) = self.paikka
        (vx, vy) = vxvy
        (x, y) = (x + vx*Dt, y + vy*Dt)
        # Lisäämällä tähän väliin muutaman if lauseen, voit korjata ohjelmaa
        # niin, että Lätkä ei karkaa peli-ikkunan reunan ulkopuolelle
        # jos x on suurempi kuin niin Wx x.stä kannattaa vähentää Wx jne...
        if x < 0:
            x = Wx
        if x > Wx:
            x = 0
        if y < 0:
            y = Wy
        if y > Wy:
            y = 0
        self.paikka = (x, y)
        self.liikkuu = True


# hyttysen väriin laitetaan satunnainen määrä punaista, vihreää ja sinistä
# Värin määrä on kokonaisluku väliltä [0,255]
class Cl_Hyttynen:
    def __init__(self, x, y):
        self.paikka = (x, y)
        self.nopeus = (70.0, 50.0)  # kokeile tähän eri arvoja
        self.koko = 10
        self.color = (random.randint(0, 255),
                      random.randint(0, 255),
                      random.randint(0, 255))
        self.elossa = True

# Piirretään hyttynen
# Elävä hyttynen on pyöreä, liiskattu hyttynen litteä ja punainen
    def piirra(self):
        iPaikka = (int(self.paikka[0]), int(self.paikka[1]))
        iEllxy = (iPaikka[0], iPaikka[1], 4*self.koko, self.koko)
        if self.elossa:
            pygame.draw.circle(screen, self.color, iPaikka, self.koko, 0)
        else:
            pygame.draw.ellipse(screen, self.color, iEllxy, 0)

# Hyttysen liike riippuu siitä, on se elossa vai kuollut
    def liiku(self):
        self.isku()  # Onko lätkä osunut?
        if self.elossa:
            self.nopeus = self.lenna(self.paikka, self.nopeus)
        else:
            self.nopeus = self.putoa()
# Siirrytään eteenpäin nopeus kertaa aikaväli
        self.paikka = dfdt(self.paikka, self.nopeus)

#   Jos Latka liikkuu lähellä tätä hyttystä sekä x- että y-suunnassa
#   lätkä osuu, niin hyttynen kuolee -- litistyy ja muuttuu punaiseksi.
    def isku(self):
        if (etaisyys(self.paikka, Latka.paikka) < Latka.koko and Latka.liikkuu):
            self.elossa = False
            self.color = Pun

# Liiskatun hyttysen nopeus.
# Vajoaa hitaasti vähän vaakasuunnassa heilahdellen.
# Satunnaisluvut saavat putoamisen näyttämään luonnollisemmalta.
    def putoa(self):
        vx = random.uniform(-10, 10)
        if self.paikka[1] < Wy-self.koko:
            vy = random.uniform(20, 70)
        else:
            vx = 0
            vy = 0
        return (vx, vy)

# Elossa olevan hyttysen kiihtyvyys on satunnaisluku.
# Lento muistuttaa Brownin liikettä.
    def lenna(self, paikka, nopeus):
        kiihtyvyys = (random.uniform(-MaxKiihtyvyys, MaxKiihtyvyys),
                      random.uniform(-MaxKiihtyvyys, MaxKiihtyvyys))
        nopeus = reunat(paikka, dfdt(nopeus, kiihtyvyys))
        return nopeus

# Edelliset funktiot olivat olion hyttynen metodeja,
# koska sisennys määrittää ne olion sisäisiksi

# Seuraavatkin voisivat olla, mutta ovat nyt yleisiä, 'kaikkien käytettävissä'
# olevia funktioita


# matka = nopeus*aika
def dfdt(xy, dxdy):
    (x, y) = xy
    (dx, dy) = dxdy
    return (x + dx*Dt, y + dy*Dt)


# Laitetaan hyttynen 'pomppamaan' takaisin näytön reunasta.
def reunat(xy, vxvy):
    (vx, vy) = vxvy
    (x, y) = xy
    # ei päästetä hyttystä reunojen ulkopuolelle
    if (x > Wx and vx > 0) or (x < 0 and vx < 0):
        vx = -vx
    elif (y > Wy and vy > 0) or (y < 0 and vy < 0):
        vy = -vy
    # pysäytetään hyttynen, mikäli se lentää tolkuttoman lujaa
    elif sqrt(vx**2 + vy**2) > Vmax:
        vx, vy = 0.0, 0.0
    return (vx, vy)


# kahden pisteen p0 ja p1 välinen etäisyys xy-tasossa
def etaisyys(p0, p1):
    (x0, y0) = p0
    (x1, y1) = p1
    return sqrt((x1-x0)**2 + (y1-y0)**2)

# def ei_hyttysia(parvi):
#   for a in parvi:
#       if a.elossa:
#           return False
#   return True

# # # # # # # # # # # # # # # # # # # # # # # # # # #
# Pääohjelma alkaa tästä
# # # # # # # # # # # # # # # # # # # # # # # # # # #


# Pygame-peliympäristön alustaminen. Helpoimmalla pääsee, jos on valmis malli,
# Mistä näkee tällaisia tarvittavan. Dokumentaatiota on toki olemassa.
pygame.display.init()
screen = pygame.display.set_mode((0, 0), pygame.FULLSCREEN)

# Selvitetään peli-ikkunan koko näytöllä pikseleinä. Pikseli on piste näytöllä.
# Nyt saadaan näytön koko, koska peli-ikkuna on FULLSCREEN
D_Info = pygame.display.Info()
Wx = D_Info.current_w
Wy = D_Info.current_h
Wxy = (Wx, Wy)  # peli-ikkunan korkeus ja leveys

pygame.init()
clock = pygame.time.Clock()
random.seed()

# Tekstin kirjoittaminen peli-ikkunaan on pikkuisen kömpelöä pygamessa.
# Tein siksi oman pikku modulin sitä varten
# luodaan oli esittelyt
esittelyt = OhjeModuli.ClGraph("Hyttysjahti")

esittelyt.tekstit((0, 0, 0),
                  [("Liiskaa hyttyset lätkällä!", (0.02, 0.02)),
                  ("z, x: vasemmalle, oikealle", (0.02, 0.05)),
                  ("ylänuoli, alanuoli: ylös, alas", (0.02, 0.08)),
                  ("esc: exit", (0.02, 0.11))])

# Luodaan hyttyslätkä näytön oikeaan alakulmaan.
# Pygamessa, ja monessa muussakin videopelissä, y-koordinaatti kasvaa alaspäin
# eli piste (0,0) on vasemmassa yläkulmmassa ja
# (x,max, ymax) oikeassa alakulmassa
Latka = Cl_Latka(Wx-50, Wy-50, 15)

# luodaan parvi hyttysiä lähelle näytön vasenta yläkulmaa
# parvi esitetään python-listana
# Listan voi tehdä kahdella tavalla
hyttyset = [Cl_Hyttynen(50, 50) for i in range(N_hyttyset)]
# hyttyset = list(range(N_hyttyset))
# for i in range(N_hyttyset):
#    hyttyset[i] = Cl_Hyttynen(50, 50)

# Toistetaan allaolevaa pelisilmukkaa, kunnes käyttäjä sulkee peli-ikkunan
# painamalla Esc-iä
# On tärkeää, että pelisilmukassa asiat tehdään oikeassa järjestyksessä.

loppu = False
while not loppu:

    # Odotetaan, että edellisestä clock.tick() kutsusta on kulunut
    # 1/Nayttotaajuus sekuntia
    clock.tick(Nayttotaajuus)

    # Tarkistetaan, onko käyttäjä käynyt painamassa esc-näppäintä tai jotenkin
    # sulkenut peli-ikkunan
    for event in pygame.event.get():
        if event.type == pygame.QUIT:
            loppu = True
        if event.type == pygame.KEYDOWN and event.key == pygame.K_ESCAPE:
            loppu = True

    # Aletaan piirtää uuttaa kuvaa, joka lopuksi läväytetään näytölle
    # Täytetään kuva taustavärillä
    screen.fill(Tausta)

    # Lisätään kuvaan ohjetekstit
    esittelyt.blitTekstit(screen)

    # 'nollataan' tieto lätkän liikkeestä.
    # Lätkä liikkuu ja tappaa vain, jos käyttäjä sitä liikuttaa
    Latka.liikkuu = False

#   Tarkistetaan, onko joku tai jotkin lätkää ohjaavista näppäimistä
# painettuna. Annetaan lätkälle nopeus vaaka- ja pystysuuntaan
# liiku() metodi laittaa todeksi muuttujan Latka.liikkuu
    pressed = pygame.key.get_pressed()
    if pressed[pygame.K_UP]:
        Latka.liiku((0.0, -VLatka))  # ylös
    if pressed[pygame.K_DOWN]:
        Latka.liiku((0.0, VLatka))  # alas
    if pressed[pygame.K_z]:
        Latka.liiku((-VLatka, 0.0))  # vasemmalle
    if pressed[pygame.K_x]:
        Latka.liiku((VLatka, 0.0))  # oikealle

    # Piirretään lätkä uudelle paikalleen
    Latka.piirra()

#  Lasketaan hyttysille uudet paikat ja piirretään ne
    for i in range(N_hyttyset):
        hyttyset[i].liiku()
        hyttyset[i].piirra()

    # näytetään kaikki yllä 'piirretty' näytöllä
    # eli flipataan uusi kuva vanhan tilalle näyttömuistiin
    pygame.display.flip()

    # while silmukka päättyy tähän
# Ohjelman loppu
pygame.quit()
