# -*- coding: utf-8 -*-
import pygame
import random
from math import pi, cos, sin, sqrt
import OhjeModuli

Musta = (0, 0, 0)  # Mustassa ei ole mitään valoa
Sin = (0, 0, 255)  # vain sinistä
Pun = (255, 0, 0)
Vihr = (0, 255, 0)
Kelt = (255, 255, 0)
Valk = (255, 255, 255)  # valkoisessa on kaikenväristä valoa
Tausta = (160, 200, 255)
MaastoVari = (95, 95, 20)

# Peli-ikkuna päivitetään Nayttotaajuus kertaa sekunnissa
Nayttotaajuus = 48

# Tällä tykin säätö saadaan jouhevaksi kaikilla näyttötaajuuksilla
saatoaskel = 48.0/Nayttotaajuus

# Mörssärillä ammutaan muutaman kilometrin päähän.
# Peli on liian hidas, ellei lentoa nopeuta.
# Erotin nopeutuksen näyttötaajuudesta ja simulointiaskeleesta, että
# se tulisi havainnollisemmin esille
# Nopeutus = 1.0 ->  oikea "filmin" nopeus
Nopeutus = 5

# Maastossa tuulee niin, että ilmanvastuksen vaikutus havainnollistuu
# paremmin
Tuuli = random.uniform(-50.0, 50.0)

g_maa = 9.81  # putoamiskiihtyvyys
RKuula = 0.1  # kuulan säde metreinä


# Lasketaan, mihin tietty pelimaailman piste kuuluu peli-ikkunassa
# reaalimaailman koordinaatit:
# vasen alanurkka: (0, 0)
# oikea ylänurkka: (Wxmaasto, Wymaasto)
# peli-ikkunan koordinaatit:
# vasen alanurkka: (0, Wy)
# oikea ylänurkka: (Wx, 0)
def xy_ruudulla(xmaasto, ymaasto):
    xpikseli = int(xmaasto*scale)
    ypikseli = int(Wy-ymaasto*scale)
#    print(xmaasto,ymaasto,Wy,scale)
    return (xpikseli, ypikseli)


# kappaleen leveyden ja korkeuden skaalaus
# sori, vähän epämääräinen
def xy_skaalaus(w, h):
    xpikseli = int(w*scale)
    ypikseli = int(h*scale)
    return (xpikseli, ypikseli)


# koordinaatistomuunnos
def r_xy(kulma, r):
    return (r*cos(kulma), r*sin(kulma))


def kuulanMassa(tiheys):
    VKuula = 4/3*pi*RKuula**3
    return tiheys*VKuula


# Kuulan kiihtyvyys, kun ilmanvastus otetaan huomioon
# Ilmanvastus oletettu verrannolliseksi nopeuden neliöön
def kiihtyvyys(vx, vy, A, m):
    c = 0.15
    vx = vx - Tuuli  # Kuulan vaakanopeus ilman suhteen
    v2 = vx**2 + vy**2
    v = sqrt(v2)
    Fd = -c*A*v2/m
    ax = vx/v*Fd
    ay = vy/v*Fd-g_maa
    return (ax, ay)


# Siirtymä ja nopeuden muutos ajassa dt.
# Numeerinen integrointi Eulerin algoritmilla
def integ_euler(x, y, vx, vy, A, M):
    dt = 1.0/Nayttotaajuus
    for i in range(Nopeutus):
        (ax, ay) = kiihtyvyys(vx, vy, A, M)
        (x, y, vx, vy) = (x + vx*dt, y + vy*dt, vx + ax*dt, vy + ay*dt)
    return (x, y, vx, vy)


# Muutetaan satunnaisesti asteittain väriä.
# Käytetään räjähdysefektissä
def hehku(color):
    (r, g, b) = color
    r = max(0, min(255, r + random.randint(-10, 10)))
    g = max(0, min(255, g + random.randint(-10, 10)))
    b = max(0, min(255, b + random.randint(-10, 10)))
    return (r, g, b)


# Vuori ja "tuuliviiri"
class Cl_vuori:
    def __init__(self):
        self.x1 = random.uniform(0.3*Wxmaasto, 0.6*Wxmaasto)
        self.leveys = 0.1*Wxmaasto
        self.h = random.uniform(0.0, 0.6*Wymaasto)

    def piirra(self):
        # Vuori
        (a, b) = xy_ruudulla(self.x1, self.h)
        (w, h) = xy_skaalaus(self.leveys, self.h)
        pygame.draw.rect(screen, MaastoVari, (a, b, w, h), 0)
        pygame.draw.line(screen, MaastoVari, (0, Wy-5), (Wx, Wy-5), 10)
        # Tuuliviirin masto
        mastonKorkeus = 0.05*Wy
        pygame.draw.line(screen, Musta, (a+w/2, b),
                         (a+w/2, b-mastonKorkeus), 4)
        # Tuuliviiri
        tuulix = Tuuli/50.0*0.05*Wx
        pygame.draw.line(screen, Sin, (a+w/2, b-mastonKorkeus),
                         (a+w/2+tuulix, b-mastonKorkeus), 12)

    # Tarkistetaan, onko kuula, jonka koordinaatit ovat x ja y osunut vuoreen
    # Pygamessa on collision detection funktioita, joita voi käyttää
    # monimutkaisemmissa tapauksissa.
    # Näin on opettavaisempi ja selkeämpi
    def osuma(self, x, y):
        return y < self.h and self.x1 < x and x < self.x1 + self.leveys


#
# ===========================================================================
# Maali
# ===========================================================================
class Cl_maali:
    def __init__(self, xmin):
        self.x = random.uniform(xmin, 0.95*Wxmaasto)
        self.y = 0.0   # 0.035*Wymaasto
        self.leveys = 0.05*Wxmaasto
        self.korkeus = 0.035*Wymaasto
        (x, y) = xy_ruudulla(self.x, self.korkeus+self.y)
        (w, h) = xy_skaalaus(self.leveys, self.korkeus)
        self.maali = pygame.Rect(x, y, w, h)
        self.vari = (225, 255, 0)
        self.ehja = True

    def piirra(self):
        if self.ehja:
            pygame.draw.ellipse(screen, self.vari, self.maali, 0)

    def osuma(self, x, y):
        if (self.ehja and y < (self.korkeus+self.y) and
                self.x < x < self.x + self.leveys):
            self.ehja = False
            for i in range(500):
                sirpaleet.append(Cl_sirpale(self.x, self.korkeus, self.vari))
            osuma_snd.play()
            return True
        else:
            return False


# Maali räjähtää erikokoisiksi ja värisiksi sirpaleiksi, jotka
# jotka lähtevät lentämään satunnaisella nopeudella
# viuhkamaisesti maalin keskipisteestä
class Cl_sirpale:
    def __init__(self, x, y, color):
        self.x = x
        self.y = y
        self.sx = self.x*random.uniform(2, 12)/100.0
        self.sy = self.y*random.uniform(2, 12)/100.0
        self.paksuus = self.y*random.uniform(2, 12)/100.0
        self.A = self.sx*self.sy
        self.M = self.A*self.paksuus*random.uniform(500.0, 2000.0)
        self.color = color
        (self.vx, self.vy) = r_xy(random.uniform(0.5, pi-0.5),
                                  random.uniform(10.0, 150.0))

    def piirra(self):
        (x, y) = xy_ruudulla(self.x, self.y)
        self.sx = min(12, max(2, self.sx + random.randint(-2, 2)))
        self.sy = min(12, max(2, self.sy + random.randint(-2, 2)))
        self.color = hehku(self.color)
        pygame.draw.rect(screen, self.color,
                         (x+int(self.sx/2), y+int(self.sy/2),
                          int(self.sx), int(self.sy)), 0)

    # Sirpaleet liikkuvat samalla dynamiikalla kuin kuulatkin
    def liiku(self):
        if self.y > 30.0:
            (self.x, self.y, self.vx, self.vy) = integ_euler(
                        self.x, self.y, self.vx, self.vy, self.A, self.M)


#
# ============================================================================
# Morssari
# ============================================================================
# Mörssärin vaunu on kummallinen pygame.Rect-olio, joka on kätevä, kunhan
# jaksaa siihen perehtyä

# Mörssäriä voi liikutella
# Ajopanoksen ruutimäärä voi vaihdella
# Kuulan tiheys voi vaihdella
# Mörssärin lataaminen kestää satunnaisen ajan
# __init__() antaa alkuarvot
class Cl_morssari:
    def __init__(self):
        self.x = 0.07*Wxmaasto
        self.y = 0.02*Wymaasto
        (w, h) = xy_skaalaus(0.05*Wxmaasto, 0.04*Wymaasto)
        self.w = w
        self.h = h
        self.ruuti = 1.0
        self.KuulanTiheys = 5000.0
        self.kuulanVari = Sin
        self.latausaika = Nayttotaajuus
        self.kulma = pi/7.0  # piipun asento
        self.p1x = 0.0  # piipun pään x-koordinaatti
        self.p1y = 0.0  # piipun pään y-koordinaatti
        self.vaunu = pygame.Rect(0, 0, w, h)

    def piirra(self):
        (x, y) = xy_ruudulla(self.x, self.y)
        # piirretään mörssärin piippu
        (sx, sy) = r_xy(self.kulma, 200.0)
        self.p1x = self.x + sx
        self.p1y = self.y + sy
        # viiveet voi toteuttaa näin vähentämällä laskuria kullakin
        # pelisilmukan kierroksella
        # (piirrä ohjelmaa kutsutaan joka pelisilmukan kierroksella)
        if self.latausaika > 0:
            self.latausaika = self.latausaika - 1
            vari = Musta
        else:
            # Piippu muuttuu punaiseksi, kun mörssäri on laukaisuvalmis
            vari = Pun
        pygame.draw.line(screen, vari, (x, y),
                         xy_ruudulla(self.p1x, self.p1y), 10)

        # Ajopanosten ruutimäärän osoitin
        pygame.draw.line(screen, Sin, (10, Wy),
                         (10, Wy-int(self.ruuti/12.0*Wy)), 8)
        # vaunu
        self.vaunu.center = (x-self.w/4, y)
        pygame.draw.ellipse(screen, Musta, self.vaunu, 0)

    # Vaunun liikuttelua z ja x näppäimillä.
    def liiku_x(self, dx):
        self.x = self.x + dx
        # Vaunulla ei pääse vuoren läpi ;-)
        if self.x > vuori.x1:
            self.x = 0.0

# num+ ja num- näppäimillä kutsutaan tätä metodia
# kun halutaan suurentaa tai pienentää ruutipanosta
    def add_ruuti(self, m):
        self.ruuti = min(11.0, max(1.0, self.ruuti + 0.5*m))
        self.ruuti = min(11.0, max(1.0, self.ruuti + 2.0*m))
# num-1, num-2 ja num-3 näppäimillä valitaan kuulan tyyppi

    def valitseKuula(self, tyyppi):
        if tyyppi == 'puu':
            self.KuulanTiheys = 1000.0
            self.kuulanVari = Kelt
        elif tyyppi == 'rauta':
            self.KuulanTiheys = 5000.0
            self.kuulanVari = Sin
        elif tyyppi == 'lyijy':
            self.KuulanTiheys = 12000.0
            self.kuulanVari = Musta

# Käännetään piippua nuolinäppäimillä
    def asento(self, dkulma):
        self.kulma = self.kulma + dkulma

# laukaistaan mörssäri välilyöntinäppäimellä
# Luodaan uusi olio ammus piipun suulle ja annetaan sille alkunopeus
    def ammu(self):
        # lähtölaukauksen ääni
        morssari_snd.play()
        # seuraavan ammuksen lataukseen kuluva aika
        self.latausaika = random.randint(Nayttotaajuus/2, 2*Nayttotaajuus)
        M = kuulanMassa(self.KuulanTiheys)
        (vx, vy) = self.kuulan_alkunopeus(M)
        return Cl_ammus(self.p1x, self.p1y, vx, vy, M, self.kuulanVari)

    # Kuulan alkunopeus riippuu ruutipanoksen suuruudesta
    def kuulan_alkunopeus(self, M):
        Cv0 = 600.0
        v0 = Cv0*sqrt(self.ruuti/M)
        print("ruuti: {:2.1f}".format(self.ruuti),
              "kulma: {:2.1f}".format(self.kulma*180.0/pi),
              "massa: {:4.1f}".format(M),
              "nopeus: {:4.0f}".format(v0))
        return (v0*cos(self.kulma),
                v0*sin(self.kulma))


# ==========================================================================
# Ammus
# ==========================================================================
class Cl_ammus:
    def __init__(self, x, y, vx, vy, M, vari):
        self.x = x
        self.y = y
        self.vx = vx
        self.vy = vy
        self.M = M  # kg/m**3
        self.A = pi*RKuula**2
        self.vari = vari
        self.rajahdys = False
        self.rajahdysvaihe = 0
        self.rajahtanyt = 0

    def piirra(self):
        if self.rajahdys:
            vari = Pun
        else:
            vari = self.vari
        if not self.rajahtanyt:
            pygame.draw.circle(screen, vari,
                               xy_ruudulla(self.x, self.y), 5, 0)
            (a, b) = xy_ruudulla(self.x, self.y)
            # Piirretään ammukselle pyrstö, jonka pituus riippuu
            # ammuksen lentonopeudesta
            (c, d) = xy_ruudulla(self.x-self.vx, self.y-self.vy)
            pygame.draw.line(screen, Kelt, (a, b), (c, d), 2)

    def liiku(self):
        # Onko osunut vuoreen, maahan tai maaliin
        self.rajahdys = (vuori.osuma(self.x, self.y) or (self.y < 20) or
                         maali.osuma(self.x, self.y))
        if not self.rajahdys:
            (self.x, self.y, self.vx, self.vy) = integ_euler(
                    self.x, self.y, self.vx, self.vy, self.A, self.M)
        else:
            self.vx = 0.0
            self.vy = 0.0
            self.rajahda()

    # räjähdyksen vaiheet
    # 0: ääni
    # (0, Nayttotaajuus/2): säkenöintiä
    # Nayttotaajuus/2: räjähtänyt
    def rajahda(self):
        if self.rajahdysvaihe < Nayttotaajuus:
            if self.rajahdysvaihe == 0:
                kuula_snd.play()
                kuula_snd.fadeout(600)
            self.rajahdysvaihe = self.rajahdysvaihe + 1
            for i in range(200):
                # säkenöinti ei noudata fysiikan lakeja
                # piirretään satunnaisesti värikkäitä viivoja
                # sitä pidempiä, mitä pidemmälle räjähdys on edennyt
                (p0x, p0y) = xy_ruudulla(self.x, self.y)
                (dx, dy) = r_xy(random.uniform(0, 2*pi),
                                self.rajahdysvaihe*random.uniform(0.25, 2.5))
                p1x = p0x + int(dx)
                p1y = p0y - int(dy)
                vari = (random.randint(160, 255),
                        random.randint(80, 255), 0)
                pygame.draw.line(screen, vari, (p0x, p0y), (p1x, p1y), 1)
        else:
            self.rajahtanyt = True


# ===========================================================================
# Pääohjelma alkaa tästä
# ==========================================================================

pygame.init()

pygame.mixer.init(frequency=22050, size=-16, channels=2, buffer=512)
osuma_snd = pygame.mixer.Sound(
    'Explosion_Ultra_Bass-Mark_DiAngelo-1810420658.wav')
morssari_snd = pygame.mixer.Sound('morssari.wav')
kuula_snd = pygame.mixer.Sound('kuula.wav')

pygame.display.init()
screen = pygame.display.set_mode((0, 0), pygame.FULLSCREEN)

# Selvitetään peli-ikkunan koko näytöllä pikseleinä
D_Info = pygame.display.Info()
Wx = D_Info.current_w
Wy = D_Info.current_h

# pelimaailman leveys "metreinä"
Wxmaasto = 5000.0
scale = Wx/Wxmaasto

# Korkeus pelimaailmassa
Wymaasto = Wy/scale

clock = pygame.time.Clock()
random.seed()

esittelyt = OhjeModuli.ClGraph("Mörssäri")
esittelyt.tekstit((0, 0, 0),
      [("Ammu kuula maaliin!", (0.02, 0.02)),
      ("z, x: siirrä tykkiä vasemmalle, oikealle", (0.02, 0.05)),
      ("ylänuoli, alanuoli: käännä putkea", (0.02, 0.08)),
      ("numeronäppäimistö +, -: lisää/vähennä ruutia", (0.02, 0.11)),
       ("numeronäppäimistö 1, 2, 3: puu, rauta tai lyijykuula", (0.02, 0.14)),
       ("välilyönti: laukaise", (0.02, 0.17)),
       ("F1: toggle lentoradat", (0.02, 0.20)),
       ("F2: toggle ohjeet", (0.02, 0.23)),
       ("esc: exit", (0.02, 0.26))])

screen.fill(Tausta)

# Luodaan mörssäri ja vuori
vuori = Cl_vuori()
morssari = Cl_morssari()
kuulat = []

maali = Cl_maali(vuori.x1+2.0*vuori.leveys)

sirpaleet = []

loppu = False
PiirraLentoradat = False
NaytaOhje = True

while not loppu:

    # odotellaan, että edellisestä kerrasta on kulunut 1/Nayttotaajuus sekuntia
    clock.tick(Nayttotaajuus)

    if not PiirraLentoradat:
        # ei piirretä vanhan päälle
        screen.fill(Tausta)

    if NaytaOhje:
        esittelyt.blitTekstit(screen)

# Tarkistetaan, onko käyttäjä painanut jotain näppäintä
    for event in pygame.event.get():
        # Lopetetaan peli sulkemalla ikkuna tai painamalla ESC
        if event.type == pygame.QUIT:
            loppu = True
        if event.type == pygame.KEYDOWN:
            if event.key == pygame.K_ESCAPE:
                loppu = True
            # Pyyhitäänkö peli-ikkuna puhtaaksi joka kierroksella vai ei
            if event.key == pygame.K_F1:
                PiirraLentoradat = not PiirraLentoradat
            # Näytetäänkö ohjeet
            if event.key == pygame.K_F2:
                NaytaOhje = not NaytaOhje
            # säädellään ajopanoksen ruutimäärää
            if event.key == pygame.K_KP_PLUS:
                morssari.add_ruuti(1)
            if event.key == pygame.K_KP_MINUS:
                morssari.add_ruuti(-1)
            # valitaan kuulan tyyppi
            if event.key == pygame.K_KP1:
                morssari.valitseKuula('puu')
            if event.key == pygame.K_KP2:
                morssari.valitseKuula('rauta')
            if event.key == pygame.K_KP3:
                morssari.valitseKuula('lyijy')

# Tarkistetaan, onko joku näppäin pohjassa juuri nyt
    # Suunnataan ja siirrellään mörssäriä
    pressed = pygame.key.get_pressed()
    if pressed[pygame.K_UP]:
        morssari.asento(0.005*saatoaskel)
    if pressed[pygame.K_DOWN]:
        morssari.asento(-0.005*saatoaskel)
    if pressed[pygame.K_x]:
        morssari.liiku_x(saatoaskel)
    if pressed[pygame.K_z]:
        morssari.liiku_x(-saatoaskel)
    # Laukaistaan mörssäri, lisätään kuula kuulien listaan
    if pressed[pygame.K_SPACE] and morssari.latausaika == 0:
        kuulat.append(morssari.ammu())

    # Piirretään kaikki oliot peli-ikkunaan ja lasketaan kuulille ja
    # sirpaleille uudet paikat peli-ikkunassa
    vuori.piirra()
    morssari.piirra()

    if maali.ehja:
        maali.piirra()

    # piirretään kuulat
    if not kuulat == []:
        for k, kuula in enumerate(kuulat):
            kuula.liiku()
            kuula.piirra()
            if kuula.rajahtanyt:
                #räjähtäneet kuulat poistetaan listalta
                del kuulat[k]

    if not sirpaleet == []:
        for sirpale in sirpaleet:
            sirpale.liiku()
            sirpale.piirra()
            # sirpaleita ei siivota, jäävät lojumaan maahan

    pygame.display.flip()
    # while silmukka päättyy tähän

pygame.quit()
