# -*- coding: utf-8 -*-
'''
<sp>
<weblink id="rk4-wikip" 
webref="http://en.wikipedia.org/wiki/Runge%E2%80%93Kutta_methods">
Runge-Kutta integration 
</weblink>
<sp>
'''
#shd 0-0:m
'''
Numeerinen integrointi Runge-Kutta menetelmällä
'''
#cod 0-0:m 1-100:z
#lasketaan y(t+dt), kun dy(t)/dt = f(y(t),u(t))
# argumentit:
# dxdt:  funktio f, palauttaa tilamuuttujien derivaatat
# yy: y(t), järjestelmän tilamuuttujat
# uu: u(t), ulkoiset ohjaukset ja parametrit
# dt: integrointiaskel

# funktio palauttaa listan yy1, joka on y(t+dt)

# algoritmi: http://en.wikipedia.org/wiki/Runge%E2%80%93Kutta_methods

def rk4(fdydt, yy, uu, dt):
    kk1 = fdydt(yy, uu)
    yk1 = [yy[i] + k1 * dt / 2.0 for i, k1 in enumerate(kk1)]
    kk2 = fdydt(yk1, uu)
    yk2 = [yy[i] + k2 * dt / 2.0 for i, k2 in enumerate(kk2)]
    kk3 = fdydt(yk2, uu)
    yk3 = [yy[i] + k3 * dt for i, k3 in enumerate(kk3)]
    kk4 = fdydt(yk3, uu)
    yy1 = [
        y + dt / 6.0 * (kk1[i] + 2.0 * kk2[i] + 2.0 * kk3[i] + kk4[i])
        for i, y in enumerate(yy)
    ]
    return yy1

#def rk4(dxdt, yy, uu, dt):
#    dim = len(yy)  # järjestelmän dimesio, tilamuuttujien lukumäärä
#    yk1 = [0.0 for i in range(dim)]
#    yk2 = [0.0 for i in range(dim)]
#    yk3 = [0.0 for i in range(dim)]
#    yy1 = [0.0 for i in range(dim)]
#    kk1 = dxdt(yy, uu)
#    for i in range(dim):
#        yk1[i] = yy[i]+kk1[i]*dt/2.0
#    kk2 = dxdt(yk1, uu)
#    for i in range(dim):
#        yk2[i] = yy[i]+kk2[i]*dt/2.0
#    kk3 = dxdt(yk2, uu)
#    for i in range(dim):
#        yk3[i] = yy[i]+kk3[i]*dt
#    kk4 = dxdt(yk3, uu)
#    for i in range(dim):
#        yy1[i] = yy[i] + dt/6.0*(kk1[i] + 2.0*kk2[i] + 2.0*kk3[i] + kk4[i])
#    return yy1
